#!/usr/bin/env python3

import argparse
import hashlib
import os

from collections import namedtuple
from datetime import datetime
from pathlib import Path, PureWindowsPath

from metadata_db import read_metadata, MetaDB


IGNORE_PROGRAM_NAMES = {'README', 'HELP', 'MANUAL', '__INSTH', 'SOUND',
                        'NOSOUND', 'PRINTME', 'UVCONFIG', 'FIX', 'PATCH',
                        'ARJ', 'FINISH', 'CONNECT'}
# Order matters: items to the left have more priority
RUNNABLE_EXTENSIONS = ['BAT', 'EXE', 'COM']
SETUP_PROGRAM_NAMES = ['SETSOUND', 'SNDSETUP', 'SOUNDSET', 'SETUP', 'INSTALL']

XML_ENTRY_FIELD_MAP = [('Name', 'name'),
                       ('ReleaseYear', 'year'),
                       ('Developer', 'developer'),
                       ('Publisher', 'publisher'),
                       ('Genres', 'genre')]

File = namedtuple('File', ['name', 'extension', 'size'])


class GameEntry:
    def __init__(self):
        self.id = None
        self.name = None
        self.genre = None
        self.series = None
        self.developer = None
        self.publisher = None
        self.year = None
        self.status = None
        self.favorite = False
        self.game_executable = None
        self.run_cmds = []
        self.config_cmds = []

    def is_blank(self):
        return self.id is None and self.name is None

    def serialize(self):
        """Serialize this entry to a string in the LAUNCHBX.TXT format."""
        lines = [self.name]

        if self.genre:
            lines.append(f'---GENRE---{self.genre}')

        if self.series:
            lines.append(f'---SERIES---{self.series}')

        if self.developer:
            lines.append(f'---DEVELOPER---{self.developer}')

        if self.publisher:
            lines.append(f'---PUBLISHER---{self.publisher}')

        if self.status:
            lines.append(f'---STATUS---{self.status}')

        if self.favorite:
            lines.append(f'---FAVORITE---')

        if self.year:
            lines.append(f'---YEAR---{self.year}')
        else:
            lines.append('---YEAR---    ')

        lines.extend(self.run_cmds)

        if self.config_cmds:
            lines.append('---CONFIG---')
            lines.extend(self.config_cmds)

        return '\r\n'.join(lines) + '\r\n\r\n'

    def __str__(self):
        return self.name


def is_configuration_program(file_: File):
    return file_.name in SETUP_PROGRAM_NAMES


def read_launchbx_file(path: Path):
    entries = {}

    game_entry = GameEntry()
    in_config_block = False

    with open(path, 'r') as f:
        for line in f.readlines():
            line = line.strip()
            if line == '':
                for run_cmd in game_entry.run_cmds:
                    if run_cmd.lower().startswith("cd "):
                        game_entry.id = run_cmd[run_cmd.rfind('\\')+1:].upper()
                        break

                if game_entry.id is None:
                    raise Exception('Failed to set ID for game: %s', game_entry.__dict__)

                # Build the DOS path to the game's executable by 'executing'
                # the steps of its run commands, in order to determine the name
                # of the executable
                if game_entry.run_cmds:
                    dos_path = PureWindowsPath()
                    for run_cmd in game_entry.run_cmds:
                        run_cmd = run_cmd.lower()

                        if run_cmd.startswith('cd'):
                            dos_path /= run_cmd[3:]
                        else:
                            dos_path /= run_cmd

                    # Assume the extension of the executable is '.exe' if it
                    # isn't specified in the run command
                    if dos_path.suffix == '':
                        game_entry.game_executable = dos_path.name + '.exe'
                    else:
                        game_entry.game_executable = dos_path.name

                entries[game_entry.id] = game_entry
                game_entry = GameEntry()
                in_config_block = False
                continue

            if not line.startswith('-'):
                if game_entry.name is None:
                    game_entry.name = line
                elif in_config_block:
                    game_entry.config_cmds.append(line)
                else:
                    game_entry.run_cmds.append(line)

                continue

            if 'CONFIG' in line:
                in_config_block = True
                continue
            else:
                in_config_block = False

            if 'GENRE' in line:
                game_entry.genre = line[11:]
            elif 'SERIES' in line:
                game_entry.series = line[12:]
            elif 'DEVELOPER' in line:
                game_entry.developer = line[15:]
            elif 'PUBLISHER' in line:
                game_entry.publisher = line[15:]
            elif 'YEAR' in line:
                game_entry.year = line[10:]
            elif 'STATUS' in line:
                game_entry.status = line[11:]
            elif 'FAVORITE' in line:
                game_entry.favorite = True

        # The entries file should have ended with a newline, so all games
        # should've been added to the dictionary and game_entry is a newly instantiated, blank object
        assert game_entry.is_blank()

    return entries


def add_games_from_dir(games_dir: Path, dos_games_dir: str, launchbox_entries):
    """Find entries in the games directory and add them """
    # Format: [(directory name, game program, setup program),]
    found_entries = []

    dos_drive = dos_games_dir[:2].lower()
    dos_games_dir = dos_games_dir[2:].lower()

    for entry in os.scandir(games_dir):
        if not os.path.isdir(entry.path):
            continue

        # Select the most likely files to run and setup the game.
        #
        # Rules for game executables:
        #  * If runnable matches the rules for a setup program (see below), it loses
        #  * In all cases: if runnable matches name of directory, it wins
        #  * BAT files have priority
        #    * In case there are multiple BAT files: largest one wins
        #  * EXE have priority over COM files
        #    * In case there are multiple: largest one wins
        #
        # Rules for setup programs:
        #  * Matches the following names, in this order: setsound, sndsetup, setup, install

        game_files = os.scandir(entry.path)
        dir_name = entry.name.upper()

        if dir_name in launchbox_entries:
            continue

        best_game_candidate = None
        best_config_candidate = None
        for file_ in game_files:
            (filename, extension) = os.path.splitext(file_.name)
            file_ = File(filename.upper(), extension[1:].upper(), file_.stat().st_size)

            if file_.extension not in RUNNABLE_EXTENSIONS:
                continue

            if file_.name in IGNORE_PROGRAM_NAMES:
                continue

            if is_configuration_program(file_):
                # Program names earlier in the list have priority
                if best_config_candidate is None or \
                        SETUP_PROGRAM_NAMES.index(file_.name) < SETUP_PROGRAM_NAMES.index(best_config_candidate.name):
                    best_config_candidate = file_
                continue

            if best_game_candidate is None:
                best_game_candidate = file_
                continue

            # In all cases: if runnable matches name of directory, it wins
            if file_.name == dir_name:
                best_game_candidate = file_
                break

            # Extensions earlier in the list win over later ones
            if RUNNABLE_EXTENSIONS.index(file_.extension) < RUNNABLE_EXTENSIONS.index(best_game_candidate.extension):
                best_game_candidate = file_
                continue

            # If the current best candidate has the same extension, the larger one wins
            if file_.extension == best_game_candidate.extension:
                if file_.size > best_game_candidate.size:
                    best_game_candidate = file_
                continue

        if best_game_candidate is None:
            print('WARNING: Could not find matching runnable file for', entry.name)
            continue

        print(dir_name, '->', best_game_candidate, best_config_candidate)
        found_entries.append((dir_name, best_game_candidate, best_config_candidate))

        game_entry = GameEntry()
        game_entry.id = dir_name
        game_entry.name = dir_name
        game_entry.game_executable = best_game_candidate.name + '.' + best_game_candidate.extension
        game_entry.run_cmds = [
            dos_drive,
            'cd ' + dos_games_dir + '\\' + dir_name.lower(),
            game_entry.game_executable,
        ]
        if best_config_candidate:
            game_entry.config_cmds = [
                dos_drive,
                'cd ' + dos_games_dir + '\\' + dir_name.lower(),
                best_config_candidate.name + '.' + best_config_candidate.extension,
            ]

        launchbox_entries[dir_name] = game_entry


def enrich_entries(game_entries, local_games_dir: Path, metadata_db: MetaDB):
    for game_entry in game_entries.values():
        try:
            with open(local_games_dir / game_entry.id / game_entry.game_executable, 'rb') as f:
                hasher = hashlib.new('md5')
                hasher.update(f.read())
                md5_hash = hasher.hexdigest()
        except FileNotFoundError:
            print(f'Cannot find path for {game_entry}:', local_games_dir / game_entry.id / game_entry.game_executable)
            continue

        xml_element = metadata_db.lookup(game_entry.name, game_entry.game_executable, md5_hash)
        if xml_element is not None:
            old_name = game_entry.name
            if enrich_game_entry(game_entry, xml_element):
                print('Enriched', game_entry)
            if old_name != game_entry.name:
                print(old_name, '->', '', end='')
                print(game_entry.name)


def enrich_game_entry(game_entry, metadata_item) -> bool:
    enriched_entry = False

    for xml_field, entry_field in XML_ENTRY_FIELD_MAP:
        if metadata_item.find(xml_field) is not None and metadata_item.find(xml_field).text is not None:
            xml_value = metadata_item.find(xml_field).text.strip()
            if getattr(game_entry, entry_field) != xml_value:
                setattr(game_entry, entry_field, xml_value)
                enriched_entry = True

    # Special case: sometimes the release date is given in full (date + time)
    if not game_entry.year and metadata_item.find('ReleaseDate') is not None:
        release_date = datetime.strptime(metadata_item.find('ReleaseDate').text, '%Y-%m-%dT%H:%M:%S%z')
        game_entry.year = release_date.year
        enriched_entry = True

    return enriched_entry


def write_launchbx_file(path: Path, launchbox_entries):
    entries = list(launchbox_entries.values())
    entries.sort(key=lambda i: i.name)

    with open(path, 'w') as f:
        for entry in entries:
            f.write(entry.serialize())


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Adds (and updates) games to LaunchBox')
    parser.add_argument("games_paths", metavar="LOCAL:DOS", type=str,
                        help='Path mapping from the games directory on the local machine, to the '
                        'same path in DOS(Box) (e.g. "/home/user/DOS/games:C:\\\\games")')
    parser.add_argument("metadata_xml", metavar="Metadata.xml", type=Path,
                        help='Path to LaunchBox\'s Metadata.xml (see download_metadata.sh)")')
    parser.add_argument("launchbx_txt", metavar="Launchbx.txt", type=Path,
                        help='Path to LaunchBox\'s LAUNCHBX.TXT")')
    args = parser.parse_args()

    entries = {}
    print('Loading Metadata XML...')
    metadata_db = read_metadata(args.metadata_xml)
    local_path, dos_path = args.games_paths.split(':', maxsplit=1)
    local_games_dir = Path(os.path.expanduser(local_path))

    launchbx_txt = Path(os.path.expanduser(args.launchbx_txt))
    print(f'Reading {launchbx_txt}...')
    if os.path.exists(launchbx_txt):
        entries = read_launchbx_file(launchbx_txt)
    print(f'Adding games from {local_games_dir}...')
    add_games_from_dir(local_games_dir, dos_path, entries)
    print('Enriching entries...')
    enrich_entries(entries, local_games_dir, metadata_db)

    print(f'Writing {launchbx_txt}...')
    write_launchbx_file(launchbx_txt, entries)
