from pathlib import Path
from xml.etree import ElementTree


class MetaDB:
    def __init__(self):
        # Format: [normalized name] => xml_element
        self.by_name = {}
        # Format: [executable name] => [xml_element, ...]
        self.by_executable = {}
        # Format: [md5] => xml_element
        self.by_md5 = {}

    def add(self, xml_element):
        self.by_name[normalize_name(xml_element.find('Name').text)] = xml_element

        if xml_element.find('StartupFile') is not None:
            self.by_executable \
                .setdefault(normalize_executable(xml_element.find('StartupFile').text), []) \
                .append(xml_element)

        if xml_element.find('StartupMD5') is not None:
            self.by_md5[normalize_md5(xml_element.find('StartupMD5').text)] = xml_element

    def lookup(self, name: str, executable_name: str, md5_hash: str):
        name = normalize_name(name)
        executable_name = normalize_executable(executable_name)
        md5_hash = normalize_md5(md5_hash)

        if name in self.by_name:
            return self.by_name[name]

        if md5_hash in self.by_md5:
            return self.by_md5[md5_hash]

        if executable_name in self.by_executable:
            if len(self.by_executable[executable_name]) == 1:
                return self.by_executable[executable_name][0]

        return None

    def __len__(self):
        return len(self.by_name)


def read_metadata(xml_path: Path) -> MetaDB:
    """
    Reads the given Launchbox Metadata XML file, filters out all non-DOS
    games and builds a database instance for querying the data.
    """
    db = MetaDB()

    for event in ElementTree.iterparse(xml_path, ['end']):
        (evt_type, element) = event
        if element.tag == 'Game':
            if (
                element.find('Platform').text == 'MS-DOS' or
                (element.find('DOS') is not None and element.find('DOS').text == 'true')
            ):
                db.add(element)

    return db


def normalize_name(name: str):
    if name is None:
        return ''

    # TODO Leestekens weghalen, misschien zelfs spaties
    return name.strip().lower()


def normalize_executable(executable: str):
    if executable is None:
        return ''

    return executable.strip().lower()


def normalize_md5(md5_hash: str):
    if md5_hash is None:
        return ''

    return md5_hash.upper()
